#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["DepthwiseConv2DTranposeOnnxConverter"]

import akida

from .conv2d_transpose import Conv2DTranposeOnnxConverter
from .conv_commons import parse_convolutional_parameters
from .register import register_onnx_converter_target


@register_onnx_converter_target("QuantizedDepthwise2DTranspose")
class DepthwiseConv2DTranposeOnnxConverter(Conv2DTranposeOnnxConverter):
    """Convert QuantizedDepthwise2DTranspose type node into an akida.DepthwiseConv2DTranspose.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """

    def _parse_akida_layer(self):
        layer_params = parse_convolutional_parameters(self, depthwise=True, transpose=True)
        return akida.DepthwiseConv2DTranspose(**layer_params)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["Conv2DTranposeOnnxConverter"]

import akida

from .activation import parse_activation_type
from .base_converter import OnnxConverter
from .conv_commons import (parse_convolutional_parameters, set_convolutional_variables,
                           check_convolution_compatibility)
from .padding import compute_conv_transpose_same_pads
from .register import register_onnx_converter_target


@register_onnx_converter_target("QuantizedConv2DTranspose")
class Conv2DTranposeOnnxConverter(OnnxConverter):
    """Convert QuantizedConv2DTranspose type node into an akida.Conv2DTranspose.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """

    def load_attributes(self, node):
        # Load default attributes
        super().load_attributes(node)

        # Some attributes should infer from node.op_type
        n_op = node.op_type
        self.activation = parse_activation_type(n_op)

    def _additional_checks(self):
        check_convolution_compatibility(self)

        # The only supported strides in akida is (2, 2).
        if self.strides != [2, 2]:
            raise ValueError(f"{self.name} expects to have strides equal to (2, 2), "
                             f"found '{self.strides}'.")

        # Check padding is SAME
        kernel_shapes = self.weights["W"].shape[-2:]
        expected_padding = compute_conv_transpose_same_pads(kernel_shapes, self.strides)
        if self.pads != expected_padding:
            raise ValueError(f"{self.name} expects to have '{expected_padding}' pads, "
                             f"found '{self.pads}'.")

    def _parse_akida_layer(self):
        layer_params = parse_convolutional_parameters(self, transpose=True)
        return akida.Conv2DTranspose(**layer_params)

    def _set_akida_variables(self, ak_layer):
        set_convolutional_variables(self, ak_layer, flip=False, transpose=True)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["Conv2DOnnxConverter", "InputConv2DOnnxConverter"]

import akida

from .base_converter import OnnxConverter
from .register import register_onnx_converter_target
from .conv_commons import (check_convolution_compatibility, parse_convolutional_parameters,
                           set_convolutional_variables)
from .activation import parse_activation_type


@register_onnx_converter_target("QuantizedConv2D")
class Conv2DOnnxConverter(OnnxConverter):
    """Convert QuantizedConv2D type node into an akida.Conv2D.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """

    def load_attributes(self, node):
        # Load default attributes
        super().load_attributes(node)

        # Some attributes should infer from node.op_type
        n_op = node.op_type
        self.activation = parse_activation_type(n_op)
        if "GlobalAvgPool" in n_op:
            self.pool_type = akida.PoolType.Average
        elif "MaxPool" in n_op:
            self.pool_type = akida.PoolType.Max
        else:
            self.pool_type = akida.PoolType.NoPooling

        # Padding type is inferred from pads attribute
        self.pads = self.weights["pads"].tolist()

    def _additional_checks(self):
        check_convolution_compatibility(self)

    def _parse_akida_layer(self):
        layer_params = parse_convolutional_parameters(self)
        return akida.Conv2D(**layer_params)

    def _set_akida_variables(self, ak_layer):
        set_convolutional_variables(self, ak_layer)


@register_onnx_converter_target("QuantizedInputConv2D")
class InputConv2DOnnxConverter(Conv2DOnnxConverter):
    """Convert QuantizedInputConv2D type node into an akida.InputConv2D.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """
    @property
    def is_input_layer(self):
        return True

    def _parse_akida_layer(self):
        layer_params = parse_convolutional_parameters(self)
        return akida.InputConv2D(input_shape=self.input_shape, **layer_params)

    def _set_akida_variables(self, ak_layer):
        set_convolutional_variables(self, ak_layer, flip=False)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["set_relu_variables", "set_lut_variables", "parse_activation_type"]

import akida
import numpy as np

from .weights import broadcast_and_set_variable, to_value_shift
from .scale_out import set_output_scale_variables


def set_relu_variables(ak_layer, max_value=None):
    """Set max value into akida variables.

    Args:
        ak_layer (akida.Layer): the akida layer to set variables.
        max_value (int, optional): the maximal value to set. Defaults to None.
    """
    if max_value is not None:
        # Max value is converted to a fixed point: unsigned 8bit << unsigned 8bit
        max_value, max_value_shift = to_value_shift(max_value, signed=False)
        max_value = np.array(max_value, dtype="uint8")
        max_value_shift = np.array(max_value_shift, dtype="uint8")
        broadcast_and_set_variable(ak_layer.variables, "max_value", max_value)
        broadcast_and_set_variable(ak_layer.variables, "max_value_shift", max_value_shift)


def set_lut_variables(ak_layer, act_scale, act_shift, lut_values):
    """Set activation variables into akida layer.

    Args:
        ak_layer (akida.Layer): the akida layer to set variables.
        act_scale (np.ndarray): activation scale.
        act_shift (np.ndarray): activation shift.
        lut_values (np.ndarray): activation table.
    """
    set_output_scale_variables(ak_layer, act_scale, act_shift, prefix="activation")
    # LUT_values in ONNX follow the indices format : [0, 1, ... N-1, -N, -(N-1), ..., -1]
    # But akida expects : [-N, -(N-1), ..., -1, 0, 1, ... N-1]
    N = lut_values.shape[0] // 2
    lut_values = np.concatenate([lut_values[N:], lut_values[:N]])
    broadcast_and_set_variable(ak_layer.variables, "lut_values", lut_values)


def parse_activation_type(op_type):
    """Parse activation type from op_type to akida.ActivationType.

    Args:
        op_type (str): the operator type.

    Returns:
        ActivationType: the corresponding Akida activation type.
    """
    akida_act_type = akida.ActivationType.NoActivation
    if "ReLU" in op_type:
        akida_act_type = akida.ActivationType.ReLU
    elif "LUT" in op_type:
        akida_act_type = akida.ActivationType.LUT
    return akida_act_type



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert input layers to Akida layers.
"""
import tensorflow as tf

from akida import Model, InputData
from ..akida_versions import AkidaVersion, get_akida_version


def get_tensor_dtype(keras_tensor, signed=True):
    """Returns the dtype of ``keras_tensor``

    Args:
        keras_tensor (:obj:`KerasTensor`): the Keras tensor to read the dtype from.
        signed (bool, optional): if a floating type is found, returns tf.int8 if signed=True,
            otherwise tf.uint8. Defaults to True.

    Returns:
        tf.Dtype: The input dtype.
    """
    input_dtype = keras_tensor.dtype
    if not input_dtype.is_integer:
        # Returns a default dtype following signed, in order to validate
        # quantizations made prior to QML 0.13.0
        return tf.int8 if signed else tf.uint8
    return input_dtype


def _create_input_data(keras_tensor):
    """Parses a KerasTensor and returns the corresponding Akida InputData layer.

    Args:
        keras_tensor (:obj:`KerasTensor`): the Keras tensor to convert.

    Returns:
        akida.InputData: The parsed layer.
    """
    input_shape = keras_tensor.shape
    input_dtype = get_tensor_dtype(keras_tensor)
    if input_shape[0] is not None:
        raise ValueError("First dimension for input layer expected to be None")
    if input_dtype.is_unsigned:
        raise ValueError("Only signed inputs are supported.")
    # Ignore first dimension, because it is None
    input_shape = input_shape[1:]
    # input shape must not exceeds 3 dim
    if len(input_shape) > 3:
        raise ValueError("input shape must not exceed 3 dimensions."
                         f"Receives {input_shape}")

    # Create a list with 1 for each dimension that we miss before having 3
    # dims, that is what akida expects
    missing_dimensions = [1] * (3 - len(input_shape))
    input_shape = [*missing_dimensions, *input_shape]
    # the akida version determines the input_bits
    if get_akida_version() == AkidaVersion.v1:
        input_bits = 4
    else:
        input_bits = input_dtype.size * 8
    # With the convert api of cnn2snn the model should handle only unsigned inputs.
    return InputData(input_shape=input_shape, input_bits=input_bits, name=keras_tensor.name)


def convert_input(model_ak, model_k):
    """Converts a keras Input layer into ``akida.InputData`` and add it to the
    Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the Akida InputData will be inserted.
        model_k (:obj:`keras.Layer`): the keras quantized model.
    """
    if not isinstance(model_ak, Model):
        raise TypeError(f"Expecting an akida model, received {type(model_ak)}")
    if len(model_ak.layers) != 0:
        raise TypeError("Akida model should be empty")

    # Get InputData from model input
    input_data = _create_input_data(model_k.input)
    model_ak.add(input_data)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["AddOnnxConverter"]

import numpy as np

import akida

from .base_converter import OnnxConverter
from .register import register_onnx_converter_target
from .weights import broadcast_and_set_variable
from .scale_out import set_output_scale_variables
from .activation import parse_activation_type


def set_input_shift_variable(ak_layer, shift, var_name="input_shift"):
    # Shift integrity check
    ak_shift = np.array(np.log2(shift), "uint8")
    if np.any(2.0**ak_shift != shift) or np.any(shift == 0):
        raise ValueError(f"Error found when setting shift in {ak_layer.name}:",
                         f"{shift} is not a power of two.")
    # Set variable
    broadcast_and_set_variable(ak_layer.variables, var_name, ak_shift)


@register_onnx_converter_target("QuantizedAdd")
class AddOnnxConverter(OnnxConverter):
    """Convert QuantizedAdd type node into akida.Add.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model containing the node.
    """

    def load_attributes(self, node):
        # Some attributes should infer from node.op_type
        n_op = node.op_type
        self.output_bits = 8 if "Scaled" in n_op else 32
        self.activation = parse_activation_type(n_op)
        return super().load_attributes(node)

    def _additional_checks(self):
        # Reject Add(X, X) (inputs are the same tensor)
        if self._node.input[0] == self._node.input[1]:
            raise RuntimeError(f"Impossible to convert {self.name} into akida. "
                               "Inputs must come from different tensors.")

    def _parse_akida_layer(self):
        return akida.Add(output_bits=self.output_bits, activation=self.activation, name=self.name)

    def _set_akida_variables(self, ak_layer):
        assert isinstance(ak_layer, akida.Add)

        # Input shifts
        set_input_shift_variable(ak_layer, self.weights["Xs"], "a_shift")
        set_input_shift_variable(ak_layer, self.weights["Ys"], "b_shift")

        # Shift out
        set_output_scale_variables(ak_layer, shift=self.weights.get("Shift", 1))



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np

import akida

from .base_converter import OnnxConverter
from .padding import get_akida_padding
from .weights import broadcast_and_set_variable, set_weight_variables
from .activation import set_relu_variables, set_lut_variables
from .scale_out import set_output_scale_variables


def _check_if_squared(value, name=""):
    assert hasattr(value, "__iter__")
    if value[0] != value[1]:
        raise ValueError(f"{name} are expected to be square.")


def _parse_additional_parameters(converter, layer_params):
    # Pooling
    if pool_type := getattr(converter, 'pool_type', None):
        layer_params["pool_type"] = pool_type
        if converter.pool_type == akida.PoolType.Max:
            layer_params["pool_size"] = converter.pool_size[0]
            layer_params["pool_stride"] = converter.pool_strides[0]
    # Activation
    layer_params["activation"] = converter.activation


def parse_convolutional_parameters(converter, depthwise=False, transpose=False):
    """Parse a converter into akida parameters

    Args:
        converter (OnnxConverter): the converter to extract parameters.
        depthwise (bool, optional): boolean to declare the layer is depthwise. Defaults to False.
        transpose (bool, optional): boolean to declare the layer is transposed. Defaults to False.

    Returns:
        dict: the parameters
    """
    layer_params = {
        "name": converter.name,
        "output_bits": 8,
        "kernel_size": converter.weights["W"].shape[-1]
    }
    # Add padding and kernel_stride parameter if not a Transpose convolution
    filters = converter.weights["W"].shape[1]
    if not transpose:
        layer_params["padding"] = get_akida_padding(converter)
        layer_params["kernel_stride"] = converter.strides[0]
        # Filters in not transposed layer are in first dimension
        filters = converter.weights["W"].shape[0]
    # Add filters parameters if not a Depthwise convolution
    if not depthwise:
        layer_params["filters"] = filters
    _parse_additional_parameters(converter, layer_params)
    return layer_params


def set_convolutional_variables(converter, ak_layer, flip=True, transpose=False):
    """Transfer converter weights to ak_layer.

    Args:
        converter (OnnxConverter): the converter to extract weights.
        ak_layer (akida.Layer): the target Akida model.
        flip (bool, optional): Boolean to flip W and H dimensions. Defaults to True.
        transpose (bool, optional): Boolean to transpose C and F dimensions. Defaults to False.
    """
    assert isinstance(converter, OnnxConverter)

    # Set padding value (in case of InputConv2D)
    if ak_layer.parameters.layer_type == akida.LayerType.InputConv2D:
        ak_variables = ak_layer.variables
        broadcast_and_set_variable(ak_variables, "padding_value", converter.weights["x_pad_value"])

    # Get kernel and transpose them (FCKxKy -> KxKyCF)
    kernel = converter.weights["W"].transpose((2, 3, 1, 0))
    if flip:
        # Kernel need to be flipped
        kernel = np.flip(kernel, axis=(0, 1))

    # Transpose C and F when required
    if transpose:
        kernel = kernel.transpose((0, 1, 3, 2))

    bias = converter.weights.get("bias", None)
    set_weight_variables(ak_layer, kernel, bias)

    # Activation
    if converter.activation == akida.ActivationType.ReLU:
        set_relu_variables(ak_layer, converter.weights.get("max_value", None))
    elif converter.activation == akida.ActivationType.LUT:
        set_lut_variables(ak_layer,
                          converter.weights["ActScale"],
                          converter.weights["ActShift"],
                          converter.weights["LUT_values"])

    # Scale out
    set_output_scale_variables(ak_layer, converter.weights["Scale"], converter.weights["Shift"])


def check_convolution_compatibility(converter):
    """Check convolution compatibility with Akida.

    Args:
        converter (OnnxConverter): the converter to check.
    """
    kernel_shapes = converter.weights["W"].shape[-2:]
    _check_if_squared(kernel_shapes, name=f"{converter.name} kernels")
    _check_if_squared(converter.strides, name=f"{converter.name} strides")
    if getattr(converter, "pool_type", None) == akida.PoolType.Max:
        _check_if_squared(converter.pool_size, name=f"{converter.name} pool sizes")
        _check_if_squared(converter.pool_strides, name=f"{converter.name} pool strides")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["generate_onnx_model"]
import warnings

from onnx.utils import Extractor
from onnx.helper import make_node
from onnx.checker import check_model
from copy import deepcopy

import akida

from .base_converter import BRN_OPSET
from .register import map_node_to_converter


def get_convertible_model(model):
    """Return the candidated model to be converted.

    Args:
        model (ModelProto): the model to convert.

    Returns:
        ModelProto: the model to be converted.
    """
    extractor = Extractor(model)

    # Extract submodel from InputQuantizer output until Dequantizer output
    input_names = [x.name for x in model.graph.input]
    output_names = [x.name for x in model.graph.output]
    if len(model.graph.node) > 0 and model.graph.node[0].op_type == "InputQuantizer":
        input_names = model.graph.node[0].output
    min_num_layers = 1
    for idx, node in enumerate(model.graph.node):
        if node.op_type == "Dequantizer":
            output_names = node.output
            min_num_layers = 2
            break
    model_to_convert = extractor.extract_model(input_names, output_names)

    # Check model is not empty (at least one node to convert is required)
    if len(model_to_convert.graph.node) < min_num_layers:
        raise ValueError("Model is empty or does not have any node to convert.")

    # Prints a warning message with a summary of the skipped nodes
    skip_nodes = model.graph.node[idx + 1:]
    if len(skip_nodes) > 0:
        stop_layer_msg = " at node " + model.graph.node[idx - 1].name if idx > 0 else ""
        skip_layers_summary = "___________________________________________________\n"
        skip_layers_summary += "Node (type)\n"
        skip_layers_summary += "===================================================\n"
        for node in skip_nodes:
            skip_layers_summary += node.name + " (" + node.op_type + ")\n"
        skip_layers_summary += "===================================================\n"
        warnings.warn("Conversion stops" + stop_layer_msg + " because of a dequantizer. "
                      "The end of the graph is ignored:\n" + skip_layers_summary)

    # Insert a fake InputQuantizer if first layer is not an input layer
    first_layer = map_node_to_converter(model_to_convert.graph.node[0], model_to_convert)
    if not first_layer.is_input_layer:
        fake_iq = make_node("InputQuantizer", [], input_names, domain=BRN_OPSET.domain)
        model_to_convert.graph.node.insert(0, fake_iq)

    return model_to_convert


def generate_onnx_model(model):
    """Generates an Akida model based on an ONNX quantizeml model.

    Args:
        model (obj:`onnx.ModelProto`): a ONNX model to convert.

    Returns:
        akida.Model: the generated Akida model.
    """
    # Clone model to keep the original intact
    model = deepcopy(model)

    # Model must be compatible with ONNX
    check_model(model, full_check=True)

    # Checks over model
    assert len(model.graph.input) == 1, "Unsupported mode: it must have exactly one input."
    assert len(model.graph.output) == 1, "Unsupported mode: it must have exactly one output."

    # Now create akida model and iterate nodes to convert each one in an akida layer.
    model = get_convertible_model(model)
    akida_model = akida.Model()
    for node in model.graph.node:
        converter = map_node_to_converter(node, model)
        converter.convert(akida_model)
    return akida_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDepthwiseConv2D to Akida.
"""
from quantizeml.layers import (QuantizedDepthwiseConv2D, QuantizedReLU, QuantizedMaxPool2D,
                               QuantizedActivation)
from akida import DepthwiseConv2D

from ..akida_versions import AkidaVersion
from .padding import check_conv_and_max_pool_compatibility
from .pooling import max_pool_param_checks
from .conv_common import get_layer_by_type, parse_conv_block_v2, set_conv_variables_v2
from .block_converter import BlockConverter, register_conversion_patterns
from .layer_utils import get_inbound_layers

__all__ = ["DepthwiseConvBlockConverter"]

_PATTERNS = [(QuantizedDepthwiseConv2D,), (QuantizedDepthwiseConv2D, QuantizedReLU),
             (QuantizedDepthwiseConv2D, QuantizedMaxPool2D, QuantizedReLU),
             (QuantizedDepthwiseConv2D, QuantizedActivation),
             (QuantizedDepthwiseConv2D, QuantizedMaxPool2D, QuantizedActivation)]


def convert_depthwise_conv_block(model_ak, block):
    """Converts a depthwise convolutional block into an akida v2 DepthwiseConv2D layer.

    The expected sequence is:

    - QuantizedDepthwiseConv2D,
    - QuantizedMaxPool2D (optional),
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.

    """
    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, block[0])

    # Evaluate the depthwise convolutional layer parameters
    conv_params = parse_conv_block_v2(block, depthwise=True)

    # Create Akida layer
    dw_conv_ak = DepthwiseConv2D(**conv_params)
    # Add layer to the model to build its internal variables
    model_ak.add(dw_conv_ak, inbound_layers_ak)

    # Set base variables
    set_conv_variables_v2(dw_conv_ak, block)

    return True


class DepthwiseConvBlockConverter(BlockConverter):
    """Main class that should be used to check if the depthwise conv block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida DepthwiseConv2D
    layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._depthwise_conv_additional_checks()

    def _depthwise_conv_additional_checks(self):
        depth_conv = self._block[0]
        # Make sure the DepthwiseConv2D kernel size and stride params are square
        assert depth_conv.kernel_size[0] == depth_conv.kernel_size[1], (
            "DepthwiseConv2D kernel should be square")
        assert depth_conv.strides[0] == depth_conv.strides[1], (
            "DepthwiseConv2D strides should be the same on both dimensions")

        # The only weight bitwidth supported is [4, 8]
        weight_bits = depth_conv.weight_quantizer.bitwidth
        assert weight_bits in [4, 8], ("DepthwiseConv2D layer can only handle weights"
                                       f" with 4 or 8 bits. Received: {weight_bits}.")

        # Check optional pooling compatibility
        pool_layer = get_layer_by_type(self._block, QuantizedMaxPool2D)
        if pool_layer:
            check_conv_and_max_pool_compatibility(depth_conv, pool_layer)
            max_pool_param_checks(pool_layer)

    def convert(self, model_ak):
        return convert_depthwise_conv_block(model_ak, self._block)


# Register the valid depthwise conv block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, DepthwiseConvBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

""" Tools to convert a group of Akida layers in a model towards a VitEncoderBlock layer. """

__all__ = ["fold_vitencoderblock"]

import os
from tempfile import TemporaryDirectory
from akida import Model, LayerType, ActivationType, VitEncoderBlock


def _get_attention_layers(model):
    attentions = []
    for layer in model.layers:
        if layer.parameters.layer_type == LayerType.Attention:
            attentions.append(layer)
    return attentions


def _get_outbounds(target_layer, model):
    outbounds = []
    for layer in model.layers:
        if target_layer in layer.inbounds:
            outbounds.append(layer)
    return outbounds


def _validate_head_topology(block_output, model):
    # Validate optional topology: BN + ExtractToken(s) and Dense head(s)
    batch_norm = None
    extract_head, head = None, None
    extract_dist_head, dist_head, add_heads = None, None, None

    # block_output can be the inbound of batch_norm
    outbounds = _get_outbounds(block_output, model)
    if len(outbounds) == 1 and outbounds[0].parameters.layer_type == LayerType.BatchNormalization:
        batch_norm = outbounds[0]
        # batch_norm should have 1 or 2 ExtractToken outbounds
        outbounds = _get_outbounds(batch_norm, model)
        if (len(outbounds) in [1, 2]
                and all(bound.parameters.layer_type == LayerType.ExtractToken
                        for bound in outbounds)):
            if len(outbounds) == 1:
                extract_head = outbounds[0]
                # extract_head can be the inbound of a Dense2D head
                outbounds = _get_outbounds(extract_head, model)
                if len(outbounds) == 1 and outbounds[0].parameters.layer_type == LayerType.Dense1D:
                    head = outbounds[0]
            else:
                # Find which ExtractToken is the base head and which is the dist head
                if (outbounds[0].parameters.begin == 0 and outbounds[0].parameters.end == 1
                        and outbounds[1].parameters.begin == 1
                        and outbounds[1].parameters.end == 2):
                    extract_head = outbounds[0]
                    extract_dist_head = outbounds[1]
                elif (outbounds[1].parameters.begin == 0 and outbounds[1].parameters.end == 1
                        and outbounds[0].parameters.begin == 1
                        and outbounds[0].parameters.end == 2):
                    extract_head = outbounds[1]
                    extract_dist_head = outbounds[0]
                else:
                    # Invalidate batch_norm
                    batch_norm = None
                if extract_head is not None:
                    extract_head_outb = _get_outbounds(extract_head, model)
                    extract_dist_head_outb = _get_outbounds(extract_dist_head, model)
                    # Check for head and dist_head Dense2D
                    if (len(extract_head_outb) == 1
                            and extract_head_outb[0].parameters.layer_type == LayerType.Dense1D
                            and len(extract_dist_head_outb) == 1
                            and (extract_dist_head_outb[0].parameters.layer_type
                                 == LayerType.Dense1D)):
                        head = extract_head_outb[0]
                        dist_head = extract_dist_head_outb[0]
                        # Check for add_heads
                        head_outbounds = _get_outbounds(head, model)
                        dist_head_outbounds = _get_outbounds(dist_head, model)
                        if (len(head_outbounds) == 1 and len(dist_head_outbounds) == 1
                                and head_outbounds[0].parameters.layer_type == LayerType.Add
                                and head_outbounds[0] == dist_head_outbounds[0]):
                            add_heads = head_outbounds[0]
                        else:
                            # Invalidate both heads
                            head, dist_head = None, None
        else:
            # Invalidate batch_norm
            batch_norm = None
    return {'batch_norm': batch_norm,
            'extract_head': extract_head, 'head': head,
            'extract_dist_head': extract_dist_head, 'dist_head': dist_head, 'add_heads': add_heads}


def _validate_topology(attention, model):
    """ This checks that topology is aligned with a VitEncoderBlock.

    The attention inbounds and outbounds are parse and for each bound, the types and number of each
    inbound and outbound is check to match the expected.

    Args:
        attention (Layer): Attention layer to start from
        model (Model): reference model, used to retrieve layers outbounds

    Returns:
        bool or dict: False if no valid topology is found and a dict with VitEncoderBlock topology
            compliant layers
    """
    # 1. Check layers before attention
    # attention inbounds should be query, key and value Dense2D layers
    if (any([not inbound.parameters.layer_type == LayerType.Dense2D
            for inbound in attention.inbounds])):
        return False

    # query, key and value inbounds should have the same norm_mha MadNorm inbound and a single
    # outbound because only self-attention is handled for now
    query, key, value = attention.inbounds
    if (query.inbounds != key.inbounds != value.inbounds
            or not query.inbounds[0].parameters.layer_type == LayerType.MadNorm
            or not len(_get_outbounds(query, model)) == 1
            or not len(_get_outbounds(key, model)) == 1
            or not len(_get_outbounds(value, model)) == 1):
        return False
    norm_mha = query.inbounds[0]

    # norm_mha should have a single outbound
    if len(_get_outbounds(norm_mha, model)) != 3:
        return False
    block_inbound = norm_mha.inbounds[0]

    # 2. Check layers after attention
    # attention is the inbound of a Dense2D projection layer
    outbounds = _get_outbounds(attention, model)
    if len(outbounds) != 1 or not outbounds[0].parameters.layer_type == LayerType.Dense2D:
        return False
    attention_projection = outbounds[0]

    # attention_projection is the first inbound of the first Add layer.
    # The second inbound of Add is the norm_mha inbound otherwise known as the block input.
    # Note that the order of Add inbounds is also important and checked.
    outbounds = _get_outbounds(attention_projection, model)
    if (len(outbounds) != 1 or not outbounds[0].parameters.layer_type == LayerType.Add
            or attention_projection not in outbounds[0].inbounds
            or block_inbound not in outbounds[0].inbounds):
        return False
    skip_connection_1 = outbounds[0]

    # First skip connection is the inbound of both the second Add and of norm_mlp MadNorm layers
    outbounds = _get_outbounds(skip_connection_1, model)
    if (len(outbounds) != 2
            or not any(bound.parameters.layer_type == LayerType.Add for bound in outbounds)
            or not any(bound.parameters.layer_type == LayerType.MadNorm for bound in outbounds)):
        return False

    # skip_connection_2 is further validated later on
    norm_mlp = outbounds[0]
    skip_connection_2 = outbounds[1]
    if outbounds[0].parameters.layer_type == LayerType.Add:
        norm_mlp, skip_connection_2 = skip_connection_2, norm_mlp

    # norm_mlp is the inbound of Dense2D mlp_1
    outbounds = _get_outbounds(norm_mlp, model)
    if len(outbounds) != 1 or not outbounds[0].parameters.layer_type == LayerType.Dense2D:
        return False
    mlp_1 = outbounds[0]

    # mlp_1 is the inbound to Dense2D mlp_2
    outbounds = _get_outbounds(mlp_1, model)
    if len(outbounds) != 1 or not outbounds[0].parameters.layer_type == LayerType.Dense2D:
        return False
    mlp_2 = outbounds[0]

    # mlp_2 is the inbound to skip_connection_2
    outbounds = _get_outbounds(mlp_2, model)
    if len(outbounds) != 1 or outbounds[0] != skip_connection_2:
        return False

    # Further checks on skip_connection_2
    if (skip_connection_1 not in skip_connection_2.inbounds
            or mlp_2 not in skip_connection_2.inbounds):
        return False

    # Topology is validated, build a candidate
    candidate = {'block_inbound': block_inbound,
                 'norm_mha': norm_mha,
                 'query': query, 'key': key, 'value': value,
                 'attention': attention, 'attention_projection': attention_projection,
                 'skip_connection_1': skip_connection_1,
                 'norm_mlp': norm_mlp, 'mlp_1': mlp_1, 'mlp_2': mlp_2,
                 'skip_connection_2': skip_connection_2}

    # 3. Check for optional head layers
    candidate.update(_validate_head_topology(skip_connection_2, model))
    return candidate


def _validate_candidate(candidate):
    # Consider the candidate valid and invalidate as the checks progress
    ret = True

    # Expected values in the block
    hidden_size = 192
    num_heads = 3
    mlp_1_units = 768
    head_bits = 28

    # Initialize other parameters with norm_mha value, later on the checks will ensure a single
    # value throughout the block
    norm_mha = candidate['norm_mha']
    post_op_buffer_bits = norm_mha.parameters.post_op_buffer_bits
    output_bits = norm_mha.parameters.output_bits

    def _check_out_params(layer, dense=False):
        # All output_bits and post_op_buffer_bits should be the same in a block
        return (layer.parameters.output_bits == output_bits
                and layer.parameters.post_op_buffer_bits == post_op_buffer_bits)

    def _check_norm(layer):
        return (layer.input_dims[0] == 1
                and layer.input_dims[2] == hidden_size
                and _check_out_params(layer))

    def _check_dense(layer, units=hidden_size, activation=ActivationType.NoActivation):
        return (layer.parameters.units == units and
                layer.parameters.activation == activation.value and
                _check_out_params(layer))

    # Check block input shape: must be (1, num_non_patch_token + num_token, hidden_size)
    ret &= _check_norm(norm_mha)

    # Check Query, Key, Value: units must be hidden_size
    ret &= _check_dense(candidate['query'])
    ret &= _check_dense(candidate['key'])
    ret &= _check_dense(candidate['value'])

    # Attention
    ret &= candidate['attention'].parameters.num_heads == num_heads
    ret &= _check_out_params(candidate['attention'])

    # Attention projection
    ret &= _check_dense(candidate['attention_projection'])

    # First skip connection
    ret &= _check_out_params(candidate['skip_connection_2'])

    # NormMLP
    ret &= _check_norm(candidate['norm_mlp'])

    # MLP
    ret &= _check_dense(candidate['mlp_1'], units=mlp_1_units, activation=ActivationType.ReLU)
    ret &= _check_dense(candidate['mlp_2'])

    # Second skip connection
    ret &= _check_out_params(candidate['skip_connection_2'])

    # Head
    bn = candidate['batch_norm']
    if bn is not None:
        ret &= (bn.parameters.activation == ActivationType.NoActivation.value and
                _check_out_params(bn))

    head = candidate['head']
    if head is not None:
        ret &= head.parameters.activation == ActivationType.NoActivation.value

        dist_head = candidate['dist_head']
        if dist_head is None:
            ret &= (head.parameters.buffer_bits == head_bits
                    and head.parameters.output_bits == head_bits)
        else:
            ret &= (head.parameters.buffer_bits == head_bits
                    and head.parameters.output_bits == output_bits)
            ret &= (dist_head.parameters.activation == ActivationType.NoActivation.value
                    and dist_head.parameters.buffer_bits == head_bits
                    and dist_head.parameters.output_bits == output_bits)
    return ret


def _set_block_variables(block, candidate):
    for key, ly in candidate.items():
        # Skip 'block_inbound' as it is only here for topology purpose and layers that were not set
        # (optional head layers)
        if key == 'block_inbound' or ly is None:
            continue
        for var in ly.variables.names:
            # 'key' is matching the block 'prefix' for block variables names
            block.set_variable(key + "_" + var, ly.variables[var])

    def _check_and_invert(skip_name, expected):
        if candidate[skip_name].inbounds != expected:
            a_shift = block.get_variable(skip_name + '_a_shift')
            b_shift = block.get_variable(skip_name + '_b_shift')
            block.set_variable(skip_name + '_a_shift', b_shift)
            block.set_variable(skip_name + '_b_shift', a_shift)

    # Skip connection inputs order matters when it comes to variables and because VitEncoderBlock
    # always assumes a specific order for skip connections, it must be checked here
    _check_and_invert('skip_connection_1',
                      [candidate['attention_projection'], candidate['block_inbound']])
    _check_and_invert('skip_connection_2', [candidate['skip_connection_1'], candidate['mlp_2']])
    if candidate['add_heads'] is not None:
        _check_and_invert('add_heads', [candidate['head'], candidate['dist_head']])


def fold_vitencoderblock(original_model):
    """ Look for Attention layer and try to rebuild a VitEncoderBlock by parsing the topology,
    shapes and parameters.

    Layers that can be replaced by a VitEncoderBlock are:

        - MADNorm MHA
        - Query, Key, Value
        - Attention
        - Attention projection
        - First skip connection (input and attention projection addition)
        - MADNorm MLP
        - MLP1 + ReLU
        - MLP2
        - Second skip connection (first skip connection and MLP2 addition)
        - Optional BatchNormalization + ExtractToken
        - Optional Dense heads

    Args:
        original_model (Model): the model to check

    Returns:
        Model: the model with VitEncoderBlock layers when possible, the original model otherwise
    """
    # As there is no API to deep copy layers or model, the original model is duplicated by saving
    # it temporarily, which allows to preserve it and work on the duplicate
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, 'tmp_model.fbz')
        original_model.save(path)
        model = Model(path)

    # Attention is the characteristic component of a VitEncoderBlock, then retrieve them first
    attentions = _get_attention_layers(model)
    if len(attentions) == 0:
        return original_model

    vit_block_candidates = {}
    for attention in attentions:
        # Parse the model topology around the attention layer and try matching with the expected
        candidate = _validate_topology(attention, model)
        # Check candidate shapes and parameters
        if candidate and _validate_candidate(candidate):
            # Build the target block
            tokens_to_extract = 0
            if candidate['extract_head'] is not None:
                tokens_to_extract += (candidate['extract_head'].parameters.end -
                                      candidate['extract_head'].parameters.begin)
                if candidate['extract_dist_head'] is not None:
                    tokens_to_extract += (candidate['extract_dist_head'].parameters.end -
                                          candidate['extract_dist_head'].parameters.begin)

            block = VitEncoderBlock(
                hidden_size=candidate['norm_mha'].input_dims[-1],
                mlp_dim=candidate['mlp_1'].parameters.units,
                num_heads=candidate['attention'].parameters.num_heads,
                num_classes=0 if candidate['head'] is None else candidate['head'].parameters.units,
                tokens_to_extract=tokens_to_extract,
                output_bits=candidate['norm_mha'].parameters.output_bits,
                buffer_bits=candidate['norm_mha'].parameters.buffer_bits,
                post_op_buffer_bits=candidate['norm_mha'].parameters.post_op_buffer_bits)
            vit_block_candidates[block] = candidate

    if len(vit_block_candidates) == 0:
        return original_model

    # Get the model layers list and replace the candidate ones with the block
    all_layers = model.layers
    for block, candidate in vit_block_candidates.items():
        # Insert the block before the first layer of the candidate layers
        all_layers.insert(all_layers.index(candidate['norm_mha']), block)
        # Pop 'block_inbound' from candidate to prevent filtering it
        candidate_copy = candidate.copy()
        del candidate_copy['block_inbound']
        # Remove the candidate layers from all_layers
        all_layers = [c for c in all_layers if c not in candidate_copy.values()]

    # Rebuild a model: layers must be added one by one to ensure they are all properly build
    block_model = Model()
    for ly in all_layers:
        inbounds = []
        # If the layer was in the model, retrieve and set same inbounds
        if ly in model.layers:
            # Get inbounds from the original model
            original_inbounds = model.get_layer(ly.name).inbounds
            for bound in original_inbounds:
                if bound in block_model.layers:
                    # Add the layer if it's in block_model
                    inbounds.append(bound)
                else:
                    # If the layer is not in block_model then it's one of the new
                    # VitEncoderBlock layers
                    for block, candidate in vit_block_candidates.items():
                        if bound in candidate.values():
                            inbounds.append(block)
                            break
        block_model.add(ly, inbound_layers=inbounds)

        # Set back variables using original_model value since the previous call to 'add' would reset
        # them
        try:
            original_layer = original_model.get_layer(ly.name)
            for var in ly.variables.names:
                ly.set_variable(var, original_layer.variables[var])
        except Exception:
            pass

    # Finally, set block variables
    for block, candidate in vit_block_candidates.items():
        _set_block_variables(block, candidate)
    return block_model
